#!/usr/bin/env python
"""
U2FsdGVkX19cV2plTMSqbpdFaJuEChbUka1NX+X4qeSMMFuTquH1Wd+vL6e2Hf4H
CAvb+R3+MDOCjdOyhgS3DNleCDqZRRJorq7lehTQzbA/pY9TMQBMyUnm63wUOjd/
EKVAggOM+jODpiFn0+FN4le8LUoOErVSZSR7L4gka/7L69wT82r90gq12H1mFNns
judHXORdh3Iud6J8UQHqSZUSFRfnd1IetIIySIIVxxCPE0DY2CS7Y/TjmEiHduzm
l4TiHrn9MlQOJY21PIjswrB9EIS/fsQsNV09BIBh6cJTu2jPULRjOpVGw+liyvkJ
I6HynL0O7XEVC61l2wiQJA==
"""

import logging
import struct
import time
import uuid

import datetime

from ..base_msg import PSBP, PSBP_PREAMBLE, crc16
from ..table import dispatch

_logger = logging.getLogger(__name__)


class Framer(object):
    """
    Framer

    The :class:`Framer` class frames bytes into PSBP messages.
    Implements the Python iterator protocol for received messages.

    Parameters
    ----------
    read : port
      Stream of bytes to read from.
    write : port
      Stream of bytes to write to.
    """

    def __init__(self, read, write, verbose=False, dispatcher=dispatch):
        self._read = read
        self._write = write
        self._verbose = verbose
        self._broken = False
        self._dispatch = dispatcher
        self._session = str(uuid.uuid4())

    def __iter__(self):
        self._broken = False
        return self

    def breakiter(self):
        """
        Break out of message iteration.
        May be called from another thread when some thread is blocked in next().
        """
        #_logger.debug('breakiter called')
        self._broken = True

    def _time(self):
        """
        Timestamp generator.

        Returns
        -------
        str : ISO 8601 format timestamp
        """
        return datetime.datetime.utcnow().isoformat()

    def next(self):
        #_logger.debug('next called')
        msg = None
        while msg is None:
            try:
                msg = self._receive()
                if self._broken:
                    raise StopIteration
            except IOError:
                raise StopIteration
        return (msg, {'time': self._time(), 'session-uid': self._session})

    def _readall(self, size):
        """
        Read until all bytes are collected.

        Parameters
        ----------
        size : int
          Number of bytes to read.
        """
        #_logger.debug('read all called')
        data = ""
        while len(data) < size:
            d = self._read(size - len(data))
            if self._broken:
                raise StopIteration
            if not d:
                # NOTE (Buro/jgross): Force a yield here to another thread. In
                # case the stream fails midstream, the spinning here causes
                # the UI thread to lock up without yielding.
                time.sleep(0)
                continue
            data += d
        return data

    def _receive(self):
        """
        Read and build PSBP message.
        """
        #_logger.debug('receive called')
        # preamble - not readall(1) to allow breaking before messages,
        # empty input
        preamble = self._read(1)
        if not preamble:
            return None
        elif ord(preamble) != PSBP_PREAMBLE:
            if self._verbose:
                _logger.debug("unhandled byte: 0x%02x" % ord(preamble))
            return None
        # hdr
        hdr = self._readall(5)
        msg_crc = crc16(hdr, crc=0xFFFF)

        reciever_address, source_address, data_length, flag, command = struct.unpack("<BBBBB", hdr)

        # data
        read_data_length = data_length - 2
        if read_data_length != 0:
            data = self._readall(read_data_length)
            msg_crc = crc16(data, msg_crc)
        else:
            data = None

        msg_crc = msg_crc ^ 0xFFFF

        # crc
        crc = self._readall(2)
        crc, = struct.unpack(">H", crc)
        if crc != msg_crc:
            _logger.debug("crc mismatch: 0x%04X 0x%04X" % (msg_crc, crc))
            return None

        if data is not None:
            msg = PSBP(receiver_address=reciever_address, source_address=source_address, data_length=data_length,
                       flag=flag,
                       command=command,
                       payload=data, crc=msg_crc)
        else:
            msg = PSBP(receiver_address=reciever_address, source_address=source_address, data_length=data_length,
                       flag=flag,
                       command=command, crc=msg_crc)

        try:
            msg = self._dispatch(msg)
        except:
            pass
        return msg

    def __call__(self, msg, **metadata):
        """
        Build and write PSBP message.

        Parameters
        ----------
        msg : PSBP message
          PSBP message to send.
        metadata : dict
          {'time': 'ISO 8601 str'} (ignored for now)
        """
        #_logger.debug('writing: {}'.format(msg))
        self._write(msg.to_binary())
